import Cachely from 'cachely';
type Versions = {
    [name: string]: string | number;
};
type Ranges = {
    [name: string]: string;
};
type SkippedVersions = {
    [name: string]: Versions;
};
type Person = {
    name: string;
    email?: string;
    url?: string;
};
/** @example https://registry.npmjs.org/docpad-plugin-eco */
interface RegistryPackageResult {
    _id: string;
    _rev: string;
    name: string;
    description: string;
    'dist-tags': {
        latest: string;
    };
    versions: {
        [version: string]: {
            name: string;
            version: string;
            description: string;
            homepage: string;
            license: string;
            keywords: string[];
            author: Person;
            maintainers: Person[];
            contributors: Person[];
            bugs: {
                url: string;
            };
            repository: {
                type: string;
                url: string;
            };
            engines: Ranges;
            main: string;
            dependencies: Versions;
            devDependencies: Ranges;
            peerDependencies: Versions;
            scripts: {
                [script: string]: string;
            };
            gitHead: string;
            _id: string;
            _shasum?: string;
            _from?: string;
            _npmVersion: string;
            _nodeVersion: string;
            _npmUser: Person;
            dist: {
                integrity?: string;
                shasum: string;
                tarball: string;
                fileCount?: number;
                unpackedSize?: number;
                'npm-signature'?: string;
            };
            directories: object;
            _npmOperationalInternal: {
                host: string;
                tmp: string;
            };
        };
    };
    readme: string;
    maintainers: Person[];
    time: {
        [version: string]: string;
        modified: string;
        created: string;
    };
    author: Person;
    repository: {
        type: string;
        url: string;
    };
    readmeFilename: string;
    users: {
        [user: string]: boolean;
    };
    homepage: string;
    keywords: string[];
    contributors: Person[];
    bugs: {
        url: string;
    };
    license: string;
    _attachments: object;
}
interface RegistryPackageResults {
    [name: string]: RegistryPackageResult;
}
interface FetchPluginsOptions {
    /** specify the dependencies that you require inside the plugin's `package.json` engines or peerDependencies */
    requirements?: Versions;
    /** specify the `package.json` dependencies that you are using to ensure compatibility */
    dependencies?: Versions;
}
interface PluginsOptions extends FetchPluginsOptions {
    /** the database result */
    database: RegistryPackageResults;
}
interface FetchPluginOptions extends FetchPluginsOptions {
    /** the name of the package to process the data for */
    name: string;
}
interface PluginOptions extends FetchPluginOptions {
    /** the database result */
    database: RegistryPackageResults;
}
interface PluginEntry {
    /** the description for the plugin */
    description: string;
    /** the homepage for the plugin */
    homepage: string;
    /** the latest version for the plugin */
    version: string;
    /** if {@link PluginOptions} `dependencies` was provided, then this is the getPlugin/fetchPlugin result */
    compatibility?: PluginCompatibilityResult;
}
interface PluginsResult {
    /** whether or not the process succeeded */
    success: boolean;
    /**  a message about success or failure */
    message: string;
    /** the processed results for each plugin, mapped by plugin name */
    plugins: {
        [name: string]: PluginEntry;
    };
}
interface PluginCompatibilityResultBase {
    /** whether or not the process succeeded */
    success: boolean;
    /** a message about success or failure */
    message: string;
}
interface PluginCompatibilityResultFailure extends PluginCompatibilityResultBase {
    success: false;
}
interface PluginCompatibilityResultSuccess extends PluginCompatibilityResultBase {
    success: true;
    /** if there were skipped versions to ensure compatibility, this is a mapping of the skipped version and the cause */
    skippedVersions: SkippedVersions;
    /** this is the latest version available for the package */
    latestVersion: string;
    /** this is the latest compatible version available for the package */
    installVersion: string;
    /** this is an array of peer dependency names that are missing and would need to be installed */
    installPeers: string[];
}
type PluginCompatibilityResult = PluginCompatibilityResultFailure | PluginCompatibilityResultSuccess;
interface PluginClerkConfig {
    /** The keyword that is common among the packages we wish to fetch */
    keyword: string;
    /** The prefix that the names of the packages needed to begin with to be valid to our clerk */
    prefix?: string;
    /** The registry hostname including scheme */
    registryHostname: string;
}
interface PluginClerkOptions extends Partial<PluginClerkConfig> {
    /** Function to send the log messages too */
    log?: (logLevel: string, ...args: any[]) => any;
    /** The amount of milliseconds until we have to query the npm database again, defaults to one day */
    cacheDuration?: number;
}
/** Construct our PluginClerk class, setting the configuration from the options */
export default class PluginClerk {
    protected config: PluginClerkConfig;
    protected log: Function;
    protected cachely: Cachely<RegistryPackageResults>;
    protected lastPoll?: number;
    protected lastUpdate?: number;
    constructor(opts: PluginClerkOptions);
    /** Creates and returns new instance of the current class */
    static create(opts: PluginClerkOptions): PluginClerk;
    /**
     * The method that fetches the result from the registry and returns it to the cachely instance
     * @returns The fetched packages from the registry that match the keyword
     */
    protected requestDatabase({ database, offset, }?: {
        database?: RegistryPackageResults;
        offset?: number;
    }): Promise<RegistryPackageResults>;
    /** Processes the package data for a plugin name, to return relevant install information, including consideration for dependency compatibility if provided. */
    protected getPlugin({ database, name, requirements, dependencies, }: PluginOptions): PluginCompatibilityResult;
    /** Get the information for all the plugins in the database, with optional support for compatibility checks */
    protected getPlugins({ database, requirements, dependencies, }: PluginsOptions): PluginsResult;
    /**
     * Fetches the database via {@link PluginClerk#fetchDatabase} and then processes the data according to {@link PluginClerk#getPlugin}
     * @param opts - forwarded to {@link PluginClerk#getPlugin}, with `database` prefilled
     */
    fetchPlugin(opts: FetchPluginOptions): Promise<PluginCompatibilityResult>;
    /**
     * Fetches the database via {@link PluginClerk#fetchDatabase} and then processes the data according to {@link PluginClerk#getPlugins}
     * @param opts - forwarded to {@link PluginClerk#getPlugins}, with `database` prefilled
     */
    fetchPlugins(opts: FetchPluginsOptions): Promise<PluginsResult>;
    /** Fetches the unprocessed data for the plugins that were fetched from the registry via the caching layer */
    fetchDatabase(): Promise<RegistryPackageResults>;
}
export {};
//# sourceMappingURL=index.d.ts.map