"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
// Import
const cachely_1 = __importDefault(require("cachely"));
const version_range_1 = __importDefault(require("version-range"));
const wait_1 = __importDefault(require("@bevry/wait"));
/** Construct our PluginClerk class, setting the configuration from the options */
class PluginClerk {
    config;
    log;
    cachely;
    lastPoll;
    lastUpdate;
    constructor(opts) {
        if (!opts || !opts.keyword) {
            throw new Error('The plugin clerk requires a keyword to be specified, please refer to the API specification');
        }
        this.config = {
            keyword: opts.keyword,
            prefix: opts.prefix,
            registryHostname: opts.registryHostname || 'https://registry.npmjs.org',
        };
        this.log = opts.log || function () { };
        this.cachely = new cachely_1.default({
            retrieve: this.requestDatabase.bind(this),
            duration: opts.cacheDuration,
            log: opts.log,
        });
    }
    /** Creates and returns new instance of the current class */
    static create(opts) {
        return new this(opts);
    }
    /**
     * The method that fetches the result from the registry and returns it to the cachely instance
     * @returns The fetched packages from the registry that match the keyword
     */
    async requestDatabase({ database = {}, offset = 0, } = {}) {
        const me = this;
        const nameRegex = /[^a-z0-9]/;
        const url = `${me.config.registryHostname}/-/v1/search?text=keywords:${me.config.keyword}&size=250&from=${offset}`;
        let data;
        me.log('notice', 'Fetching database content');
        try {
            // fetch response, supporting retry on timeouts
            let response;
            while (true) {
                try {
                    response = await fetch(url);
                }
                catch (err) { }
                if (response && response.ok) {
                    break;
                }
                else {
                    // wait a random duration up to a minute before trying again
                    await (0, wait_1.default)(Math.random() * 60000);
                }
            }
            data = await response.json();
            if (!data || !data.objects) {
                me.log('error', 'Fetched data from', url, 'had no objects', data);
                return Promise.reject(new Error('Fetched data had no objects'));
            }
            me.log('info', 'Fetching details for', data.objects.length, 'plugins');
            await Promise.all(data.objects.map(async function (entry) {
                const name = entry.package.name;
                let pluginData;
                try {
                    const response = await fetch(`${me.config.registryHostname}/${name}`);
                    pluginData = await response.json();
                }
                catch (err) {
                    return Promise.reject(err);
                }
                if (name !== pluginData.name) {
                    return Promise.reject(new Error('name result from search and from package did not match'));
                }
                if (me.config.prefix) {
                    const code = name.replace(me.config.prefix, '');
                    if (code === name || nameRegex.test(code)) {
                        // invalid plugin
                        me.log('warn', `Plugin ${name} will be ignored as it has an invalid name, must be prefixed with: ${me.config.prefix}`);
                        return Promise.resolve();
                    }
                }
                database[name] = pluginData;
                me.log('info', `Plugin ${name} was successfully added to the database.`);
                return Promise.resolve();
            }));
        }
        catch (err) {
            me.log('error', 'Fetching database content failed', err);
            return Promise.reject(err);
        }
        // this could be optimized further
        // option 1: fetch all the pages first, then combine them, then fetch package data
        // option 2: calculate how many pages are needed, then fetch them all at once (only possible after first page is retrieved, as we need the total)
        // option 3: do both of the above
        // however, this can be a PR by someone, as need to get this out sooner than later
        const subtotal = data.objects.length + offset;
        if (subtotal + offset !== Number(data.total)) {
            me.log('info', 'Fetched a portion of the database content, grabbing the rest');
            return this.requestDatabase({ database, offset: subtotal });
        }
        me.log('notice', 'Fetched all the database content');
        return Promise.resolve(database);
    }
    /** Processes the package data for a plugin name, to return relevant install information, including consideration for dependency compatibility if provided. */
    getPlugin({ database, name, requirements = {}, dependencies = {}, }) {
        const pluginData = database[name];
        if (pluginData == null) {
            return {
                message: `The requested plugin ${name} was not found in the plugin database`,
                success: false,
            };
        }
        const skippedVersions = {}; /* {
            '2.0.0': {
                'docpad': '^6.53.0'
            }
        } */
        const latestVersion = pluginData['dist-tags'].latest;
        let installVersion;
        let installPeers = []; /* ['docpad'] */
        try {
            // cycle through the plugins versions
            // to discover the most recent version that satisfies installed dependencies
            const pluginVersionsData = pluginData.versions;
            const pluginVersionsKeysLatestFirst = Object.keys(pluginVersionsData).reverse();
            for (const pluginVersion of pluginVersionsKeysLatestFirst) {
                // skip pre-release versions
                if (pluginVersion.includes('-'))
                    continue;
                // prepare
                const pluginVersionData = pluginVersionsData[pluginVersion];
                const pluginVersionMissingPeers = [];
                // cycle through the external dependencies to check if the requirements exist
                const existingDependencies = Object.assign({}, requirements, dependencies);
                const requiredDependencies = Object.assign({}, pluginVersionData.engines || {}, pluginVersionData.peerDependencies || {});
                // ensure all required dependencies exist
                for (const [name, suppliedVersion] of Object.entries(requirements)) {
                    const acceptedRange = requiredDependencies[name];
                    if (!acceptedRange ||
                        (0, version_range_1.default)(suppliedVersion, acceptedRange) ===
                            false) {
                        // incompatibility
                        if (skippedVersions[pluginVersion] == null)
                            skippedVersions[pluginVersion] = {};
                        skippedVersions[pluginVersion][name] =
                            acceptedRange || suppliedVersion;
                    }
                }
                // ensure all peer dependencies exist
                for (const [name, acceptedRange] of Object.entries(pluginVersionData.peerDependencies || {})) {
                    const suppliedVersion = existingDependencies[name];
                    if (suppliedVersion) {
                        if ((0, version_range_1.default)(suppliedVersion, acceptedRange) === false) {
                            // incompatibility
                            if (skippedVersions[pluginVersion] == null)
                                skippedVersions[pluginVersion] = {};
                            skippedVersions[pluginVersion][name] = acceptedRange;
                        }
                    }
                    else {
                        pluginVersionMissingPeers.push(name);
                    }
                }
                // check if this version is to be skipped
                if (skippedVersions[pluginVersion]) {
                    continue;
                }
                // version is okay, so install this one
                installVersion = pluginVersion;
                installPeers = pluginVersionMissingPeers;
                break;
            }
        }
        catch (err) {
            this.log('warn', 'The compatibility checks failed with error:', err.stack);
            return {
                message: 'The compatibility checks failed',
                success: false,
            };
        }
        // Check if we found a version to install
        // @ts-ignore
        if (!installVersion) {
            return {
                message: `Failed to find a compatible version of the plugin ${name}`,
                success: false,
            };
        }
        // success
        return {
            message: `Successfully fetched ${latestVersion === installVersion ? 'the latest' : 'an older'} and compatible version of the plugin ${name}`,
            success: true,
            skippedVersions,
            latestVersion,
            installVersion,
            installPeers,
        };
    }
    /** Get the information for all the plugins in the database, with optional support for compatibility checks */
    getPlugins({ database, requirements, dependencies, }) {
        const result = {
            success: true,
            message: 'Successfully fetched the plugins',
            plugins: {},
        };
        for (const [pluginName, pluginData] of Object.entries(database)) {
            const plugin = {
                description: pluginData.description,
                homepage: pluginData.homepage,
                version: pluginData['dist-tags'].latest,
            };
            if (dependencies || requirements) {
                plugin.compatibility = this.getPlugin({
                    name: pluginName,
                    requirements,
                    dependencies,
                    database,
                });
            }
            result.plugins[pluginName] = plugin;
        }
        return result;
    }
    /**
     * Fetches the database via {@link PluginClerk#fetchDatabase} and then processes the data according to {@link PluginClerk#getPlugin}
     * @param opts - forwarded to {@link PluginClerk#getPlugin}, with `database` prefilled
     */
    async fetchPlugin(opts) {
        let database;
        try {
            database = await this.fetchDatabase();
        }
        catch (err) {
            return Promise.reject(err);
        }
        return this.getPlugin({ database, ...opts });
    }
    /**
     * Fetches the database via {@link PluginClerk#fetchDatabase} and then processes the data according to {@link PluginClerk#getPlugins}
     * @param opts - forwarded to {@link PluginClerk#getPlugins}, with `database` prefilled
     */
    async fetchPlugins(opts) {
        let database;
        try {
            database = await this.fetchDatabase();
        }
        catch (err) {
            return Promise.reject(err);
        }
        return this.getPlugins({ database, ...opts });
    }
    /** Fetches the unprocessed data for the plugins that were fetched from the registry via the caching layer */
    fetchDatabase() {
        return this.cachely.resolve();
    }
}
exports.default = PluginClerk;
